from typing import AsyncIterator, Literal
import warnings
from temporalio.client import (
    Client,
    Schedule,
    ScheduleActionStartWorkflow,
    ScheduleSpec,
    ScheduleHandle,
)
from temporalio import workflow
from temporalio.worker import Worker
from temporalio.exceptions import WorkflowAlreadyStartedError
from rojak.types import ConversationMessage
from rojak.session import Session

with workflow.unsafe.imports_passed_through():
    from rojak.retrievers import RetrieverActivities
    from rojak.agents import Agent, AgentActivities
    from rojak.workflows import (
        AgentWorkflow,
        OrchestratorWorkflow,
        OrchestratorParams,
        OrchestratorResponse,
        ShortOrchestratorParams,
        ShortOrchestratorWorkflow,
    )


class Rojak:
    def __init__(self, client: Client, task_queue: str):
        self.client: Client = client
        self.task_queue: str = task_queue

    async def create_worker(
        self,
        agent_activities: list[AgentActivities],
        retriever_activities: list[RetrieverActivities] = [],
    ) -> Worker:
        """Create a worker.

        Args:
            agent_activities (list[AgentActivities]): List of activity classes that can be called.
            retriever_activities (list[RetrieverActivities], optional): List of retriever activity classes that can be called. Defaults to [].

        Returns:
            Worker: A worker object that can be used to start the worker.
        """
        activities = []
        for activity in agent_activities:
            activities.append(activity.call)
            activities.append(activity.execute_function)
            activities.append(activity.execute_instructions)

        for retriever in retriever_activities:
            activities.append(retriever.retrieve_and_combine_results)

        worker: Worker = Worker(
            self.client,
            task_queue=self.task_queue,
            workflows=[OrchestratorWorkflow, AgentWorkflow, ShortOrchestratorWorkflow],
            activities=activities,
        )

        return worker

    async def list_scheduled_runs(
        self,
        schedule_id: str,
        statuses: list[
            Literal[
                "Running", "Completed", "Failed", "Cancelled", "Terminated", "TimedOut"
            ]
        ]
        | None = None,
        limit: int = 10,
        page_size: int = 1000,
        next_page_token: bytes | None = None,
    ) -> AsyncIterator[str]:
        """List the ID of orchestrators associated with the schedule.

        Args:
            schedule_id (str): Unique identifier of the schedule.
            statuses (list[ Literal[ 'Running', 'Completed', 'Failed', 'Cancelled', 'Terminated', 'TimedOut' ] ] | None, optional): List of statuses to filter the runs. Defaults to None.
            limit (int, optional): Maximum number of IDs to return. Defaults to 10.
            page_size (int, optional): Maximum number of results per page. Defaults to 1000.
            next_page_token (bytes | None, optional): A previously obtained next page token if doing pagination. Usually not needed as the iterator automatically starts from the beginning. Defaults to None.

        Returns:
            AsyncIterator[str]:  An async iterator that can be used with `async for`.
        """
        status_filter = (
            " OR ".join(f'ExecutionStatus="{status}"' for status in statuses)
            if statuses
            else ""
        )
        query = f'TemporalScheduledById="{schedule_id}"'
        if status_filter:
            query += f" AND ({status_filter})"

        async for workflow_execution in self.client.list_workflows(
            query=query,
            limit=limit,
            page_size=page_size,
            next_page_token=next_page_token,
        ):
            yield workflow_execution.id

    async def create_schedule(
        self,
        schedule_id: str,
        schedule_spec: ScheduleSpec,
        agent: Agent,
        messages: list[ConversationMessage],
        context_variables: dict = {},
        max_turns: int = float("inf"),
        debug: bool = False,
    ) -> ScheduleHandle:
        """Create a schedule and return its handle.

        The schedule periodically executes the equivalent of the `run()` method with the provided inputs.

        Args:
            schedule_id (str): Unique identifier of the schedule.
            schedule_spec (ScheduleSpec): Specification on when the action is taken.
            agent (Agent): The initial agent to be called.
            messages (list[ConversationMessage]): A list of message objects.
            context_variables (dict, optional): A dictionary of additional context variables, available to functions and Agent instructions. Defaults to {}.
            max_turns (int, optional): The maximum number of conversational turns allowed. Defaults to float("inf").
            debug (bool, optional): If True, enables debug logging. Defaults to False.

        Returns:
            ScheduleHandle: A handle to the created schedule.
        """
        data = ShortOrchestratorParams(
            agent, context_variables, max_turns, messages, debug
        )

        return await self.client.create_schedule(
            schedule_id,
            Schedule(
                action=ScheduleActionStartWorkflow(
                    ShortOrchestratorWorkflow.run,
                    data,
                    id=schedule_id,
                    task_queue=self.task_queue,
                ),
                spec=schedule_spec,
            ),
        )

    async def run(
        self,
        id: str,
        agent: Agent,
        messages: list[ConversationMessage],
        context_variables: dict = {},
        max_turns: int = float("inf"),
        debug: bool = False,
    ) -> OrchestratorResponse:
        """Initialise an orchestrator with the provided inputs and wait for completion.

        Requires a running worker.

        Args:
            id (str): Unique identifier of the orchestrator.
            agent (Agent): The initial agent to be called.
            messages (list[ConversationMessage]): A list of message objects.
            context_variables (dict, optional): A dictionary of additional context variables, available to functions and Agent instructions. Defaults to {}.
            max_turns (int, optional): The maximum number of conversational turns allowed. Defaults to float("inf").
            debug (bool, optional): If True, enables debug logging. Defaults to False.

        Returns:
            OrchestratorResponse: A response object containing updated messages, context_variables and agent.
        """
        data = ShortOrchestratorParams(
            agent, context_variables, max_turns, messages, debug
        )
        return await self.client.execute_workflow(
            ShortOrchestratorWorkflow.run,
            data,
            id=id,
            task_queue=self.task_queue,
        )

    async def get_run_result(self, id: str) -> OrchestratorResponse:
        """Get result from a completed orchestrator.

        Results are not stored indefinitely and may have been removed depending on your Retention Period.

        Requires a running worker.

        Args:
            id (str): Unique identifier of the orchestrator.

        Returns:
            OrchestratorResponse: A response object containing updated messages, context_variables and agent.
        """
        workflow_handle = self.client.get_workflow_handle(
            id, result_type=OrchestratorResponse
        )
        return await workflow_handle.result()

    async def get_session(self, session_id: str) -> Session:
        """Retrieve the session with the given ID.

        Args:
            session_id (str): The unique identifier for the session.

        Raises:
            ValueError: If no session with the specified ID exists.

        Returns:
            Session: The Session object associated with the given ID.
        """
        try:
            workflow_handle = self.client.get_workflow_handle(session_id)
            description = await workflow_handle.describe()
            if description.raw_info.type.name == "OrchestratorWorkflow":
                return Session(workflow_handle)
            else:
                raise
        except Exception:
            raise ValueError(
                f"Session with ID {session_id} does not exist. Please create a session first."
            )

    async def create_session(
        self,
        session_id: str,
        agent: Agent,
        context_variables: dict = {},
        max_turns: int = float("inf"),
        history_size: int = 10,
        debug: bool = False,
    ) -> Session:
        """Create a session if not yet started. The session will maintain conversation history and configurations.

        Args:
            session_id (str): Unique identifier of the session.
            agent (Agent): The initial agent to be called.
            context_variables (dict, optional): A dictionary of additional context variables, available to functions and Agent instructions. Defaults to {}.
            max_turns (int, optional): The maximum number of conversational turns allowed. Defaults to float("inf").
            history_size (int, optional): The maximum number of messages retained in the list before older messages are removed. When this limit is exceeded, the messages are summarized, and the summary becomes the first message in a new list. Defaults to 10.
            debug (bool, optional): If True, enables debug logging. Defaults to False.

        Returns:
            Session: The Session object created.
        """
        data = OrchestratorParams(
            agent=agent,
            context_variables=context_variables,
            max_turns=max_turns,
            history_size=history_size,
            debug=debug,
        )
        try:
            workflow_handle = await self.client.start_workflow(
                OrchestratorWorkflow.run,
                data,
                id=session_id,
                task_queue=self.task_queue,
            )
            return Session(workflow_handle)
        except WorkflowAlreadyStartedError:
            warnings.warn(
                "A session with this ID is already running. Returning the existing session.",
                UserWarning,
            )
            workflow_handle = self.client.get_workflow_handle(session_id)
            return Session(workflow_handle)



from temporalio.client import (
    WorkflowHandle,
)
from temporalio import workflow
from rojak.types import ConversationMessage

with workflow.unsafe.imports_passed_through():
    from rojak.agents import Agent
    from rojak.workflows import (
        OrchestratorWorkflow,
        OrchestratorResponse,
        SendMessageParams,
        UpdateConfigParams,
    )


class Session:
    def __init__(self, workflow_handle: WorkflowHandle):
        self.workflow_handle = workflow_handle

    async def send_message(
        self,
        message: ConversationMessage,
        agent: Agent,
    ) -> OrchestratorResponse:
        """Send a message to the agent specified.

        Args:
            message (ConversationMessage): New query as a message object.
            agent (Agent): Agent to send message to.
            context_variables (dict, optional): A dictionary of additional context variables, available to functions and Agent instructions. Defaults to {}.

        Returns:
            OrchestratorResponse: A response object containing updated messages, context_variables and agent.
        """
        return await self.workflow_handle.execute_update(
            OrchestratorWorkflow.send_message,
            SendMessageParams(message, agent),
            result_type=OrchestratorResponse,
        )

    async def get_result(self) -> OrchestratorResponse:
        """Get the latest response.

        Requires a running worker.

        Returns:
            OrchestratorResponse: Response object containing updated messages and context_variables.
        """
        return await self.workflow_handle.query(OrchestratorWorkflow.get_result)

    async def get_config(self) -> dict[str, any]:
        """Retrieve the current session configuration.

        Requires a running worker.

        Returns:
            dict[str, any]: A dictionary with the current session's configuration values.
        """
        return await self.workflow_handle.query(OrchestratorWorkflow.get_config)

    async def update_config(self, params: UpdateConfigParams):
        """Update the session's configuration with specified changes.

        Requires a running worker.

        Args:
            params (UpdateConfigParams): A dictionary containing only the configuration values that need to be updated.
        """
        await self.workflow_handle.signal(OrchestratorWorkflow.update_config, params)

    async def cancel(self):
        """Cancel the session."""
        return await self.workflow_handle.cancel()



import copy
from dataclasses import dataclass, field
from datetime import timedelta
from temporalio import workflow
from temporalio.exceptions import ActivityError, ChildWorkflowError
from rojak.types import (
    ConversationMessage,
    ContextVariables,
)
from collections import deque
import asyncio
from rojak.utils import create_retry_policy, debug_print
from rojak.workflows.agent_workflow import (
    AgentWorkflow,
    AgentWorkflowRunParams,
    ToolResponse,
    AgentTypes,
)
from rojak.agents import (
    Agent,
    AgentCallParams,
    AgentResponse,
)


@dataclass
class OrchestratorBaseParams:
    agent: AgentTypes
    """The initial agent to be called."""

    context_variables: ContextVariables = field(default_factory=dict)
    """A dictionary of additional context variables, available to functions and Agent instructions."""

    max_turns: int | float = field(default=float("inf"))
    """The maximum number of conversational turns allowed."""

    messages: list[ConversationMessage] = field(default_factory=list)
    """A list of message objects."""

    debug: bool = False
    """If True, enables debug logging"""


@dataclass
class OrchestratorParams(OrchestratorBaseParams):
    history_size: int = field(default=10)
    """The maximum number of messages retained in the list before older messages are removed."""


@dataclass
class ShortOrchestratorParams(OrchestratorBaseParams):
    context_variables: ContextVariables = field(default_factory=dict)
    """A dictionary of additional context variables, available to functions and Agent instructions."""


@dataclass
class OrchestratorResponse:
    """The response object from containing the updated state."""

    messages: list[ConversationMessage]
    """The list of updated messages."""

    agent: AgentTypes | None
    """The last agent to be called."""

    context_variables: ContextVariables
    """The dictionary of the updated context variables."""


@dataclass
class SendMessageParams:
    message: ConversationMessage
    """A message object."""

    agent: AgentTypes
    """The agent to be called."""


@dataclass
class UpdateConfigParams:
    context_variables: ContextVariables | None = None
    """The dictionary of the updated context variables."""

    max_turns: int | float | None = None
    """The maximum number of conversational turns allowed."""

    history_size: int | None = None
    """The maximum number of messages retained in the list before older messages are removed."""

    debug: bool | None = None
    """If True, enables debug logging"""


class OrchestratorBaseWorkflow:
    def __init__(self, params: OrchestratorBaseParams):
        self.messages: list[ConversationMessage] = params.messages
        self.max_turns = params.max_turns
        # Keep track of the last to be called
        self.agent = params.agent
        self.debug = params.debug
        self.context_variables = params.context_variables

    async def process(self, active_agent: Agent) -> Agent | None:
        response, updated_messages = await workflow.execute_child_workflow(
            AgentWorkflow.run,
            AgentWorkflowRunParams(
                agent=active_agent,
                messages=self.messages,
                context_variables=self.context_variables,
                debug=self.debug,
            ),
            execution_timeout=timedelta(minutes=1),
        )

        self.messages = updated_messages

        if isinstance(response.output, ToolResponse):
            fn_result = response.output.output
            if fn_result.agent is not None:
                debug_print(
                    self.debug,
                    workflow.now(),
                    f"{active_agent.name}: Transferred to '{fn_result.agent.name}'.",
                )
                self.agent = active_agent = fn_result.agent
            if fn_result.context_variables is not None:
                self.context_variables = fn_result.context_variables
        elif isinstance(response.output, str):
            debug_print(
                self.debug,
                workflow.now(),
                f"\n{active_agent.name}: {response.output}",
            )
            active_agent = None

        return active_agent

    @workflow.query
    def get_messages(self) -> list[ConversationMessage]:
        return self.messages


@workflow.defn
class ShortOrchestratorWorkflow(OrchestratorBaseWorkflow):
    """Orchestrator for short-running workflows."""

    @workflow.init
    def __init__(self, params: ShortOrchestratorParams) -> None:
        super().__init__(params)

    @workflow.run
    async def run(self, params: ShortOrchestratorParams) -> OrchestratorResponse:
        message = self.messages[-1]
        debug_print(self.debug, workflow.now(), f"{message.role}: {message.content}")

        active_agent = self.agent
        init_len = len(self.messages)

        while len(self.messages) - init_len < self.max_turns and active_agent:
            active_agent = await self.process(active_agent)

        return OrchestratorResponse(
            messages=self.messages,
            agent=self.agent,
            context_variables=self.context_variables,
        )


@workflow.defn
class OrchestratorWorkflow(OrchestratorBaseWorkflow):
    """Orchestrator for long-running workflows."""

    @workflow.init
    def __init__(self, params: OrchestratorParams) -> None:
        super().__init__(params)
        self.lock = asyncio.Lock()  # Prevent concurrent update handler executions
        self.queue: deque[tuple[ConversationMessage, Agent]] = deque()
        self.pending: bool = False
        self.history_size: int = params.history_size
        # Stores latest response
        self.result: OrchestratorResponse = OrchestratorResponse([], None, {})

    @workflow.run
    async def run(self, params: OrchestratorParams) -> OrchestratorResponse:
        while True:
            await workflow.wait_condition(lambda: bool(self.queue))
            message, agent = self.queue.popleft()
            past_message_state = copy.deepcopy(self.messages)
            self.messages.append(message)

            debug_print(
                self.debug, workflow.now(), f"{message.role}: {message.content}"
            )
            active_agent = self.agent = agent
            init_len = len(self.messages)

            try:
                while len(self.messages) - init_len < self.max_turns and active_agent:
                    active_agent = await self.process(active_agent)

                self.result = OrchestratorResponse(
                    messages=self.messages,
                    agent=self.agent,
                    context_variables=self.context_variables,
                )
                self.pending = False

                # Summarise chat and start new workflow if messages exceeds `history_size` limit
                if len(self.messages) > self.history_size:
                    await workflow.wait_condition(
                        lambda: workflow.all_handlers_finished()
                    )
                    response = await self.summarise()
                    debug_print(
                        self.debug, workflow.now(), f"Summary: {response.output}"
                    )
                    assert isinstance(response.output, str)
                    debug_print(
                        self.debug,
                        workflow.now(),
                        "Continue as new due to messages exceeding history size.",
                    )
                    workflow.continue_as_new(
                        args=[
                            OrchestratorParams(
                                agent=self.agent,
                                history_size=self.history_size,
                                max_turns=self.max_turns,
                                context_variables=self.context_variables,
                                messages=[
                                    ConversationMessage(
                                        role="assistant",
                                        content=response.output,
                                    )
                                ],
                                debug=self.debug,
                            )
                        ]
                    )
            except (ChildWorkflowError, ActivityError) as e:
                # Return messages to previous state and wait for new messages
                match e:
                    case ChildWorkflowError():
                        workflow.logger.error(
                            f"Failed to run agent workflow. Error: {e}"
                        )
                        self.messages = past_message_state
                        print("Revert messages to previous state.")
                    case ActivityError():
                        workflow.logger.error(
                            f"Failed to summarise messages. Error: {e}"
                        )
                    case _:
                        workflow.logger.error(f"Unexpected error. Error: {e}")
                active_agent = None
                self.pending = False
                continue

    async def summarise(self) -> AgentResponse:
        """Use the last agent used to generate a summary of the conversation."""
        messages_string = "\n\n".join(
            f"{message.role}: {message.content}" for message in self.messages
        )
        summary_prompt = (
            "This is the conversation history between a user and a chatbot:\n\n"
            + f"{messages_string}\n\n Please produce a three sentence summary of "
            + "this conversation."
        )
        response: AgentResponse = await workflow.execute_activity(
            f"{self.agent.type}_call",
            AgentCallParams(
                messages=[ConversationMessage(role="user", content=summary_prompt)],
                model=self.agent.model,
            ),
            start_to_close_timeout=timedelta(
                seconds=self.agent.retry_options.timeout_in_seconds
            ),
            retry_policy=create_retry_policy(self.agent.retry_options.retry_policy),
            result_type=AgentResponse,
        )
        return response

    @workflow.update(unfinished_policy=workflow.HandlerUnfinishedPolicy.ABANDON)
    async def send_message(
        self,
        params: SendMessageParams,
    ) -> OrchestratorResponse:
        async with self.lock:
            self.pending = True
            self.queue.append((params.message, params.agent))
            await workflow.wait_condition(lambda: self.pending is False)
            return self.result

    @workflow.query
    def get_result(self) -> OrchestratorResponse:
        return self.result

    @workflow.query
    def get_config(self) -> dict:
        return {
            "context_variables": self.context_variables,
            "max_turns": self.max_turns,
            "history_size": self.history_size,
            "debug": self.debug,
        }

    @workflow.signal
    def update_config(self, params: UpdateConfigParams):
        if params.context_variables is not None:
            self.context_variables = params.context_variables
        if params.max_turns is not None:
            self.max_turns = params.max_turns
        if params.history_size is not None:
            self.history_size = params.history_size
        if params.debug is not None:
            self.debug = params.debug
